from .triplet_loss import TripletLoss
from .kl_divergence import KL,SKL
from .symmetric_ce import SCELoss
from .focal_loss import FocalLoss
from .kd_loss import *
from .label_smoothing import LabelSmoothingCE